//
//  eprom.c	EEPROM emulation by FLASH memory
//  Copyright(C)2014 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2014.02.02 v0.0  First cording
//
//

#include <stddef.h>						// libc (NULL)
#include <string.h>						// libc (memxxx())
#include "stm32f4xx_flash.h"			// stm lib

//#define DEBUG

#ifdef DEBUG
#include <stdio.h>
#define DEBUGP(...)		printf(__VA_ARGS__)
#else
#define DEBUGP(...)
#endif

//
//	EPROM emulation
//

// Constant
const struct flashMap {
	uint16_t	FLASH_Sector;
	char		*addr;					// Linear address space
	size_t		len;
}	flashMap[] = {
		{FLASH_Sector_0,  (char *)(FLASH_BASE+0x00000), 16*1024},	// STM32F40x
		{FLASH_Sector_1,  (char *)(FLASH_BASE+0x04000), 16*1024},
		{FLASH_Sector_2,  (char *)(FLASH_BASE+0x08000), 16*1024},
		{FLASH_Sector_3,  (char *)(FLASH_BASE+0x0c000), 16*1024},
		{FLASH_Sector_4,  (char *)(FLASH_BASE+0x10000), 64*1024},
		{FLASH_Sector_5,  (char *)(FLASH_BASE+0x20000), 128*1024},
		{FLASH_Sector_6,  (char *)(FLASH_BASE+0x40000), 128*1024},
		{FLASH_Sector_7,  (char *)(FLASH_BASE+0x60000), 128*1024},
		{FLASH_Sector_8,  (char *)(FLASH_BASE+0x80000), 128*1024},
		{FLASH_Sector_9,  (char *)(FLASH_BASE+0xa0000), 128*1024},
		{FLASH_Sector_10, (char *)(FLASH_BASE+0xc0000), 128*1024},
		{FLASH_Sector_11, (char *)(FLASH_BASE+0xe0000), 128*1024},
		{0, 0, 0}
};

// Control blocks
const struct flashMap	*eprom_flashMapTop;
unsigned long			eprom_areaSize;
int 					eprom_blockNumber;
unsigned long			eprom_newEntery;	// Round robin support
unsigned long			eprom_startEntery;	// Round robin support



void eprom_open(void){
    extern char	_sidata;				/* Defined by the linker */
    extern char	_sdata;					/* Defined by the linker */
    extern char	_edata;					/* Defined by the linker */
    char		*eidata;
	const struct flashMap	*p;

	// Search the unused space on FLASH memory
    eidata = &_sidata + (&_edata - &_sdata);
    for(p = flashMap; p->len != 0; p++) {
    	if (eidata < p->addr) {
        	eprom_flashMapTop = p;
    		break;
    	}
    }

    // Initiate
    eprom_areaSize    = 0;
    eprom_blockNumber = 0;
    for(; p->len != 0; p++) {
		if (*(p->addr) != 0xff) {
			if ((p == eprom_flashMapTop) || (*(p->addr - 1) == 0xff)) {
				eprom_startEntery = eprom_areaSize;
			}
		}

    	eprom_areaSize += p->len;
    	eprom_blockNumber++;
    }
    DEBUGP("eprom_open: size=%ld blk=%d %ld -> %ld\n", eprom_areaSize, eprom_blockNumber, eprom_startEntery, eprom_newEntery);
	return;
}

void eprom_close(void){
	eprom_flashMapTop = NULL;
	eprom_areaSize    = 0;
    eprom_blockNumber = 0;
	return;
}

//unsigned long eprom_getSize(void) {
//	return(eprom_areaSize);
//}
#define eprom_getSize()			(eprom_areaSize)

//void *eprom_getMemoryAddr(unsigned long a) {
//	return(eprom_flashMapTop->addr + a);
//}
#define eprom_getMemoryAddr(a)	(void *)(eprom_flashMapTop->addr + (a))

int eprom_getBlockByAddr(unsigned long epromAddr) {
	const struct flashMap	*p;
	int						b;

	b = 0;
	for(p = eprom_flashMapTop; p->len != 0; p++) {
		if (epromAddr < p->len) {
			break;
		}
		epromAddr -= p->len;
		b++;
	}
	if (p->len) {
		return(b);
	} else {
		return(-1);
	}
}

unsigned long eprom_getAddrByBlock(int block){
	const struct flashMap	*p;
	unsigned long			a;

	a = 0;
	for(p = eprom_flashMapTop; block > 0; --block, p++) {
		a += p->len;
	}
	return(a);
}

int eprom_read(void *buf, unsigned long epromAddr, int len){
	int	rc;

	if ((epromAddr + len) <= eprom_getSize()) {
		memcpy(buf, eprom_getMemoryAddr(epromAddr), len);
		rc = 0;
	} else {
		memset(buf, 0, len);
		rc = -1;
	}
	return(rc);
}

int eprom_write(void *buf, unsigned long epromAddr, int len){	// 4bytes boundary
	int		rc;
	char	*flash;

	rc    = 0;
	flash = eprom_getMemoryAddr(epromAddr);
	if ((epromAddr + len) <= eprom_getSize()) {
		uint32_t	*f;						// From
		uint32_t	*t;						// To
		int			s;						// Write count at 32bits

		f  = (uint32_t *)buf;
		t  = (uint32_t *)flash;
		s  = len / sizeof(uint32_t);
		DEBUGP("eprom_write: %08lx data=", t);
		FLASH_Unlock();
		FLASH_ClearFlag(FLASH_FLAG_EOP | FLASH_FLAG_OPERR | FLASH_FLAG_WRPERR | FLASH_FLAG_PGAERR | FLASH_FLAG_PGPERR|FLASH_FLAG_PGSERR);
		for(; s > 0; --s, f++, t++) {
			FLASH_Status	r;

			if ((r = FLASH_ProgramWord((uint32_t)t, *f)) != FLASH_COMPLETE) {
				rc = r;
				DEBUGP("eprom_write: FLASH_ProgramWord err=%d 0x%08lx 0x%08lx 0x%08lx\n", r, t, f, FLASH->SR);
				break;
			}
			DEBUGP("%08lx ", *t);
		}
		FLASH_Lock();
		DEBUGP("\n");
	} else {
		rc = -1;
		DEBUGP("eprom_write: size err %ld %d\n", epromAddr, len);
	}
	return(rc);
}


int eprom_erase(int block){
	int						rc;
	const struct flashMap	*p;

	rc = 0;
	if ((block >= 0) && (block < eprom_blockNumber)) {
		FLASH_Status	r;

		p = eprom_flashMapTop + block;
		FLASH_Unlock();
		FLASH_ClearFlag(FLASH_FLAG_EOP | FLASH_FLAG_OPERR | FLASH_FLAG_WRPERR | FLASH_FLAG_PGAERR | FLASH_FLAG_PGPERR|FLASH_FLAG_PGSERR);
		if ((r = FLASH_EraseSector(p->FLASH_Sector, VoltageRange_3)) != FLASH_COMPLETE) {
			rc = r;
			DEBUGP("eprom_erase: FLASH_EraseSector err=%d 0x%04x 0x%08lx\n", r, p->FLASH_Sector, FLASH->SR);
		}
		FLASH_Lock();
		DEBUGP("eprom_erase: rc=%d blk=%d sec=0x%x\n", rc, block, p->FLASH_Sector);
	}
	return(rc);
}

//
//	Round robin support
//		record 1st and last byte is not 0xff
//		record size are 4byte boundary
//
void *eprom_getRoundRobin(unsigned long *control, int len) {
	void			*p;
	unsigned long	c;

	p = NULL;
	if (*control == 0) {
		*control = eprom_startEntery;
	}

	if (*control >= eprom_getSize()) {
		c = *control - eprom_getSize();
	} else {
		c = *control;
	}
	p = eprom_getMemoryAddr(c);
	if (*((unsigned char *)p) != 0xff) {
#ifdef DEBUG
		{
			int	i;
			unsigned char	*d;

			d = p;
			DEBUGP("eprom_getRoundRobin: data=");
			for(i = 0; i < len; i++){
				DEBUGP("%02x ", *d++);
			}
			DEBUGP("\n");
		}
#endif
		*control += len;
	} else {
		p = NULL;				// EOF
	}
	return(p);
}

int eprom_appendRoundRobin(void *buf, int len){	// 4bytes boundary
	int				rc;					//
	unsigned char	*flash;				//

	if ((rc = eprom_write(buf, eprom_newEntery, len)) == 0) {
		eprom_newEntery += len;
		if (eprom_newEntery >= eprom_getSize()) {
			eprom_newEntery = 0;
		}

		flash = eprom_getMemoryAddr(eprom_newEntery);
		if (*flash != 0xff) {			// Found used block
			int	b;

			b = eprom_getBlockByAddr(eprom_newEntery);
			if ((rc = eprom_erase(b)) == 0) {
				eprom_startEntery = eprom_getAddrByBlock(b+1);
				if (eprom_startEntery >= eprom_getSize()) {
					eprom_startEntery = 0;
				}
			}
			DEBUGP("eprom_appendRoundRobin: eprom_erase=%d used=%ld blk=%d new start=%ld\n", rc, eprom_newEntery, b, eprom_startEntery);
		}
	} else {
		DEBUGP("eprom_appendRoundRobin: eprom_write=%d %ld %d\n", rc, eprom_newEntery, len);
	}
	return(rc);
}

void eprom_setRoundRobinNewEntry(unsigned long addr){
	while(addr >= eprom_getSize()) {
		addr -= eprom_getSize();
	}
	eprom_newEntery = addr;
    DEBUGP("eprom_setRoundRobinNewEntry: %ld -> %ld\n", eprom_startEntery, eprom_newEntery);
	return;
}
